/* Emacs style mode select   -*- C++ -*-
 *-----------------------------------------------------------------------------
 *
 *
 *  PrBoom: a Doom port merged with LxDoom and LSDLDoom
 *  based on BOOM, a modified and improved DOOM engine
 *  Copyright (C) 1999 by
 *  id Software, Chi Hoang, Lee Killough, Jim Flynn, Rand Phares, Ty Halderman
 *  Copyright (C) 1999-2000 by
 *  Jess Haas, Nicolas Kalkhof, Colin Phipps, Florian Schulze
 *  Copyright 2005, 2006 by
 *  Florian Schulze, Colin Phipps, Neil Stevens, Andrey Budko
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 *  02111-1307, USA.
 *
 * DESCRIPTION:
 *      Startup and quit functions. Handles signals, inits the
 *      memory management, then calls D_DoomMain. Also contains
 *      I_Init which does other system-related startup stuff.
 *
 *-----------------------------------------------------------------------------
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "doomdef.h"
#include "m_argv.h"
#include "d_main.h"
#include "m_fixed.h"
#include "i_system.h"
#include "i_video.h"
#include "z_zone.h"
#include "lprintf.h"
#include "m_random.h"
#include "doomstat.h"
#include "g_game.h"
#include "m_misc.h"
#include "i_sound.h"
#include "i_main.h"
#include "r_fps.h"
#include "lprintf.h"

#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

#include <nds.h>
#include <fat.h>

/* Most of the following has been rewritten by Lee Killough
 *
 * I_GetTime
 * killough 4/13/98: Make clock rate adjustable by scale factor
 * cphipps - much made static
 */

int realtic_clock_rate = 100;
static int_64_t I_GetTime_Scale = 1<<24;

static int I_GetTime_Scaled(void)
{
  return (int)( (int_64_t) I_GetTime_RealTime() * I_GetTime_Scale >> 24);
}



static int  I_GetTime_FastDemo(void)
{
  static int fasttic;
  return fasttic++;
}



static int I_GetTime_Error(void)
{
  I_Error("I_GetTime_Error: GetTime() used before initialization");
  return 0;
}



int (*I_GetTime)(void) = I_GetTime_Error;

void I_Init(void)
{
  /* killough 4/14/98: Adjustable speedup based on realtic_clock_rate */
  if (fastdemo)
    I_GetTime = I_GetTime_FastDemo;
  else
    if (realtic_clock_rate != 100)
      {
        I_GetTime_Scale = ((int_64_t) realtic_clock_rate << 24) / 100;
        I_GetTime = I_GetTime_Scaled;
      }
    else
      I_GetTime = I_GetTime_RealTime;

  {
    /* killough 2/21/98: avoid sound initialization if no sound & no music */
    if (!(nomusicparm && nosfxparm))
      I_InitSound();
  }

  R_InitInterpolation();
}

/* cleanup handling -- killough:
 */
static void I_SignalHandler(int s)
{
  char buf[2048];

  signal(s,SIG_IGN);  /* Ignore future instances of this signal.*/

  strcpy(buf,"Exiting on signal: ");
  I_SigString(buf+strlen(buf),2000-strlen(buf),s);

  /* If corrupted memory could cause crash, dump memory
   * allocation history, which points out probable causes
   */
  if (s==SIGSEGV || s==SIGILL || s==SIGFPE)
    Z_DumpHistory(buf);

  I_Error("I_SignalHandler: %s", buf);
}



// killough 2/22/98: Add support for ENDBOOM, which is PC-specific

static void PrintVer(void)
{
  char vbuf[200];
  lprintf(LO_INFO,"%s\n",I_GetVersionString(vbuf,200));
}

static unsigned int cp437_to_unicode(unsigned char cp437)
{
  static const unsigned short cp437_to_unicode_table[256] =
  {
    0x0020, 0x263a, 0x263b, 0x2665, 0x2666, 0x2663, 0x2660, 0x2022,
    0x25d8, 0x25cb, 0x25d9, 0x2642, 0x2640, 0x266a, 0x266b, 0x263c,
    0x25ba, 0x25c4, 0x2195, 0x203c, 0x00b6, 0x00a7, 0x25ac, 0x21a8,
    0x2191, 0x2193, 0x2192, 0x2190, 0x221f, 0x2194, 0x25b2, 0x25bc,
    0x0020, 0x0021, 0x0022, 0x0023, 0x0024, 0x0025, 0x0026, 0x0027,
    0x0028, 0x0029, 0x002a, 0x002b, 0x002c, 0x002d, 0x002e, 0x002f,
    0x0030, 0x0031, 0x0032, 0x0033, 0x0034, 0x0035, 0x0036, 0x0037,
    0x0038, 0x0039, 0x003a, 0x003b, 0x003c, 0x003d, 0x003e, 0x003f,
    0x0040, 0x0041, 0x0042, 0x0043, 0x0044, 0x0045, 0x0046, 0x0047,
    0x0048, 0x0049, 0x004a, 0x004b, 0x004c, 0x004d, 0x004e, 0x004f,
    0x0050, 0x0051, 0x0052, 0x0053, 0x0054, 0x0055, 0x0056, 0x0057,
    0x0058, 0x0059, 0x005a, 0x005b, 0x005c, 0x005d, 0x005e, 0x005f,
    0x0060, 0x0061, 0x0062, 0x0063, 0x0064, 0x0065, 0x0066, 0x0067,
    0x0068, 0x0069, 0x006a, 0x006b, 0x006c, 0x006d, 0x006e, 0x006f,
    0x0070, 0x0071, 0x0072, 0x0073, 0x0074, 0x0075, 0x0076, 0x0077,
    0x0078, 0x0079, 0x007a, 0x007b, 0x007c, 0x007d, 0x007e, 0x2302,
    0x00c7, 0x00fc, 0x00e9, 0x00e2, 0x00e4, 0x00e0, 0x00e5, 0x00e7,
    0x00ea, 0x00eb, 0x00e8, 0x00ef, 0x00ee, 0x00ec, 0x00c4, 0x00c5,
    0x00c9, 0x00e6, 0x00c6, 0x00f4, 0x00f6, 0x00f2, 0x00fb, 0x00f9,
    0x00ff, 0x00d6, 0x00dc, 0x00a2, 0x00a3, 0x00a5, 0x20a7, 0x0192,
    0x00e1, 0x00ed, 0x00f3, 0x00fa, 0x00f1, 0x00d1, 0x00aa, 0x00ba,
    0x00bf, 0x2310, 0x00ac, 0x00bd, 0x00bc, 0x00a1, 0x00ab, 0x00bb,
    0x2591, 0x2592, 0x2593, 0x2502, 0x2524, 0x2561, 0x2562, 0x2556,
    0x2555, 0x2563, 0x2551, 0x2557, 0x255d, 0x255c, 0x255b, 0x2510,
    0x2514, 0x2534, 0x252c, 0x251c, 0x2500, 0x253c, 0x255e, 0x255f,
    0x255a, 0x2554, 0x2569, 0x2566, 0x2560, 0x2550, 0x256c, 0x2567,
    0x2568, 0x2564, 0x2565, 0x2559, 0x2558, 0x2552, 0x2553, 0x256b,
    0x256a, 0x2518, 0x250c, 0x2588, 0x2584, 0x258c, 0x2590, 0x2580,
    0x03b1, 0x00df, 0x0393, 0x03c0, 0x03a3, 0x03c3, 0x00b5, 0x03c4,
    0x03a6, 0x0398, 0x03a9, 0x03b4, 0x221e, 0x03c6, 0x03b5, 0x2229,
    0x2261, 0x00b1, 0x2265, 0x2264, 0x2320, 0x2321, 0x00f7, 0x2248,
    0x00b0, 0x2219, 0x00b7, 0x221a, 0x207f, 0x00b2, 0x25a0, 0x00a0,
  };
  return cp437_to_unicode_table[cp437];
}

static void unicode_to_utf8(unsigned char *utf8, unsigned int unicode)
{
  if (unicode < 0x80) {
    *utf8++ = unicode;
  } else if (unicode < 0x800) {
    *utf8++ = 0xC0 | ((unicode>>6) & 0x1F);
    *utf8++ = 0x80 | ( unicode     & 0x3F);
  } else if (unicode < 0x10000) { // only need basic plane for cp437
    *utf8++ = 0xE0 | ((unicode>>12) & 0x1F);
    *utf8++ = 0x80 | ((unicode>> 6) & 0x3F);
    *utf8++ = 0x80 | ( unicode      & 0x3F);
  }
  *utf8 = '\0';
}

static int endoom_selectlump(void)
{
  int lump_eb, lump_ed, lump = -1;

  /* CPhipps - ENDOOM/ENDBOOM selection */
  lump_eb = W_CheckNumForName("ENDBOOM");/* jff 4/1/98 sign our work    */
  lump_ed = W_CheckNumForName("ENDOOM"); /* CPhipps - also maybe ENDOOM */

  if (lump_eb == -1)
    lump = lump_ed;
  else if (lump_ed == -1)
    lump = lump_eb;
  else
  { /* Both ENDOOM and ENDBOOM are present */
#define LUMP_IS_NEW(num) (!((lumpinfo[num].source == source_iwad) || \
                            (lumpinfo[num].source == source_auto_load)))
    switch ((LUMP_IS_NEW(lump_ed) ? 1 : 0) |
            (LUMP_IS_NEW(lump_eb) ? 2 : 0)) {
    case 1:
      lump = lump_ed;
      break;
    case 2:
      lump = lump_eb;
      break;
    default:
      /* Both lumps have equal priority, both present */
      lump = (P_Random(pr_misc) & 1) ? lump_ed : lump_eb;
      break;
    }
  }

  return lump;
}

struct endoom_state
{
  int bg;
  int color;
  int bold;
  int blink;
};

static void endoom_init(struct endoom_state *state)
{
  state->bg = state->color = state->bold = state->blink = -1;
}

static void endoom_start(struct endoom_state *state)
{
#ifdef _WIN32
  printf("\n");
#else
  printf("\e[0m\n");
#endif
}

static void endoom_charstyle(struct endoom_state *state, unsigned char style)
{
#ifdef _WIN32
  I_ConTextAttr(style);
#else
  static unsigned char cga_to_ansi[] = { 0, 4, 2, 6, 1, 5, 3, 7 }; // convert()
  static unsigned char bold_to_ansi[] = { 22, 1 }; // intensity normal/high
  static unsigned char blink_to_ansi[] = { 25, 5 }; // blinking off/on

  // style is a bitfield, treat it as such
  int color = (style     ) & 0x7;
  int bold  = (style >> 3) & 0x1;
  int bg    = (style >> 4) & 0x7;
  int blink = (style >> 7) & 0x1;

  /* foreground color */
  if (color != state->color)
  {
    state->color = color;
    /* we buffer everything or output is horrendously slow */
    printf("\e[%dm", 30 + cga_to_ansi[color]);
  }

  // bold, now handled separately from foreground
  if (bold != state->bold)
  {
    state->bold = bold;
    printf("\e[%dm", bold_to_ansi[bold]);
  }

  /* background color */
  if (bg != state->bg)
  {
    state->bg = bg;
    printf("\e[%dm", 40 + cga_to_ansi[bg]);
  }

  // blink flag, if your terminal supports it
  if (blink != state->blink)
  {
    state->blink = blink;
    printf("\e[%dm", blink_to_ansi[blink]);
  }
#endif
}

static void endoom_printchar(unsigned char cp437)
{
#ifdef _WIN32
  printf("%c", cp437);
#else
  unsigned int unicode;
  unsigned char utf8[8];

  unicode = cp437_to_unicode(cp437);

  unicode_to_utf8(utf8, unicode);

  printf("%s", utf8);
#endif
}

static void endoom_endofline(struct endoom_state *state)
{
#ifdef _WIN32
#else
  // reset background colour at new line to avoid "spillage"
  state->bg = -1;
  printf("\e[49m\n");
#endif
}

static void endoom_finish(struct endoom_state *state)
{
#ifdef _WIN32
#else
  printf("\e[0m"); /* cph - reset colours */
#endif
}

/* I_EndDoom
 * Prints out ENDOOM or ENDBOOM, using some common sense to decide which.
 * cphipps - moved to l_main.c, made static
 */
static void I_EndDoom(void)
{
  int lump = endoom_selectlump();

  if (lump != -1)
  {
    const unsigned char *endoom = W_CacheLumpNum(lump);
    int i, l = W_LumpLength(lump) / 2;
    int j, n;

    /* cph - colour ENDOOM by rain */
    struct endoom_state state;
    endoom_init(&state);

    endoom_start(&state);

    for (j=0, n=80; j<l; j+=80, n+=80)
    {
      for (i=j; i<l && i<n; i++)
      {
        endoom_charstyle(&state, endoom[2*i+1]);
        endoom_printchar(endoom[2*i+0]);
      }
      endoom_endofline(&state);
    }

    endoom_finish(&state);

    W_UnlockLumpNum(lump);
  }

#ifndef _WIN32
  PrintVer();
#endif
}



static int has_exited;

/* I_SafeExit
 * This function is called instead of exit() by functions that might be called
 * during the exit process (i.e. after exit() has already been called)
 * Prevent infinitely recursive exits -- killough
 */

void I_SafeExit(int rc)
{
  if (!has_exited)    /* If it hasn't exited yet, exit now -- killough */
    {
      has_exited=rc ? 2 : 1;
      exit(rc);
    }
}

static void I_Quit (void)
{
  if (!has_exited)
    has_exited=1;   /* Prevent infinitely recursive exits -- killough */

  if (has_exited == 1) {
    I_EndDoom();
    if (demorecording)
      G_CheckDemoStatus();
    M_SaveDefaults ();
  }
}

#ifdef SECURE_UID
uid_t stored_euid = -1;
#endif

static int old_console;
static PrintConsole bottomScreen;

void keyboardStart() {
	REG_DISPCNT_SUB &= ~DISPLAY_BG3_ACTIVE;
	dmaFillWords(0,BG_GFX_SUB,128*1024);
	dmaCopy(BG_PALETTE_SUB,BG_MAP_RAM_SUB(28),512);
	keyboardInit(NULL, 0, BgType_Text4bpp, BgSize_T_256x512, 29, 1, false, true);
	BG_PALETTE_SUB[255] = RGB15(31,31,31);
	consoleSetWindow(&bottomScreen, 0,0,32,14);
	keyboardShow();
}

void keyboardEnd() {
	keyboardHide();
	consoleSetWindow(&bottomScreen, 0,0,32,24);

}



//int main(int argc, const char * const * argv)
int main(int argc, char **argv)
{
  videoSetModeSub(MODE_0_2D|DISPLAY_BG1_ACTIVE);
  vramSetBankC(VRAM_C_SUB_BG);

  REG_BG1CNT_SUB = BG_MAP_BASE(31);
  BG_PALETTE_SUB[255] = RGB15(31,31,31);

  consoleInit(&bottomScreen,1, BgType_Text4bpp, BgSize_T_256x256, 31, 0, false, true);


  powerOn(POWER_ALL);
  soundEnable();
  defaultExceptionHandler();


  TIMER0_DATA=0;	// Set up the timer
  TIMER1_DATA=0;
  TIMER0_CR=TIMER_DIV_1024 | TIMER_ENABLE;
  TIMER1_CR=TIMER_CASCADE | TIMER_ENABLE;

	// Jefklak 19/11/06 - adjust upper/lower screen stuff
	// ### UPPER SCREEN #### //
  videoSetMode(MODE_5_2D|DISPLAY_BG3_ACTIVE);	// BG3 only - extended rotation
  vramSetBankA(VRAM_A_MAIN_BG_0x06000000);		// same as VRAM_A_MAIN_BG
  vramSetBankB(VRAM_B_MAIN_BG_0x06020000);		// use second bank for main screen - 256 KiB

  REG_BG3CNT = BG_BMP8_512x512;					// BG3 Control register, 8 bits
  REG_BG3PA = 1 << 8;
  REG_BG3PB = 0; // BG SCALING X
  REG_BG3PC = 0; // BG SCALING Y
  REG_BG3PD = 1 << 8;
  REG_BG3X = 0;
  REG_BG3Y = 0;

  // clear upper screen (black) instead of junk
  memset(BG_GFX, 0, 512 * 512 * 2);
  swiWaitForVBlank();

  if (!fatInitDefault())
  {
	iprintf("Unable to initialize media device!\n");
  } else {
    iprintf("fatInitDefault(): initialized.\n");
  }

  myargc = argc;
  myargv = (const char * const *) argv;

  /* Version info */
  lprintf(LO_INFO,"\n");
  PrintVer();

  /* cph - Z_Close must be done after I_Quit, so we register it first. */
  atexit(Z_Close);
  /*
     killough 1/98:

     This fixes some problems with exit handling
     during abnormal situations.

     The old code called I_Quit() to end program,
     while now I_Quit() is installed as an exit
     handler and exit() is called to exit, either
     normally or abnormally. Seg faults are caught
     and the error handler is used, to prevent
     being left in graphics mode or having very
     loud SFX noise because the sound card is
     left in an unstable state.
  */

  Z_Init();                  /* 1/18/98 killough: start up memory stuff first */

  atexit(I_Quit);

  /* cphipps - call to video specific startup code */
  I_PreInitGraphics();

  D_DoomMain ();
  return 0;
}
